#include "tmesh.h"
#include <float.h>
#include <fstream>

void TMesh::SetCube(V3 center, float sl) {

  vertsN = 8;
  verts = new V3[vertsN];
  trisN = 12;
  tris = new unsigned int[trisN*3];

  verts[0] = center + V3(-sl/2.0f, +sl/2.0f, +sl/2.0f);
  verts[1] = center + V3(-sl/2.0f, -sl/2.0f, +sl/2.0f);
  verts[2] = center + V3(+sl/2.0f, -sl/2.0f, +sl/2.0f);
  verts[3] = center + V3(+sl/2.0f, +sl/2.0f, +sl/2.0f);
  verts[4] = center + V3(-sl/2.0f, +sl/2.0f, -sl/2.0f);
  verts[5] = center + V3(-sl/2.0f, -sl/2.0f, -sl/2.0f);
  verts[6] = center + V3(+sl/2.0f, -sl/2.0f, -sl/2.0f);
  verts[7] = center + V3(+sl/2.0f, +sl/2.0f, -sl/2.0f);

  int tri = 0;
  tris[3*tri+0] = 0;
  tris[3*tri+1] = 1;
  tris[3*tri+2] = 2;
  tri++;
  tris[3*tri+0] = 2;
  tris[3*tri+1] = 3;
  tris[3*tri+2] = 0;
  tri++;
  tris[3*tri+0] = 3;
  tris[3*tri+1] = 2;
  tris[3*tri+2] = 6;
  tri++;
  tris[3*tri+0] = 6;
  tris[3*tri+1] = 7;
  tris[3*tri+2] = 3;
  tri++;
  tris[3*tri+0] = 7;
  tris[3*tri+1] = 6;
  tris[3*tri+2] = 5;
  tri++;
  tris[3*tri+0] = 5;
  tris[3*tri+1] = 4;
  tris[3*tri+2] = 7;
  tri++;
  tris[3*tri+0] = 4;
  tris[3*tri+1] = 5;
  tris[3*tri+2] = 1;
  tri++;
  tris[3*tri+0] = 1;
  tris[3*tri+1] = 0;
  tris[3*tri+2] = 4;
  tri++;
  tris[3*tri+0] = 4;
  tris[3*tri+1] = 0;
  tris[3*tri+2] = 3;
  tri++;
  tris[3*tri+0] = 3;
  tris[3*tri+1] = 7;
  tris[3*tri+2] = 4;
  tri++;
  tris[3*tri+0] = 1;
  tris[3*tri+1] = 5;
  tris[3*tri+2] = 6;
  tri++;
  tris[3*tri+0] = 6;
  tris[3*tri+1] = 2;
  tris[3*tri+2] = 1;
  tri++;

  cols = new V3[vertsN];
  for (int i = 0; i < 4; i++) {
    cols[i] = V3(1.0f, 0.0f, 0.0f);
    cols[4+i] = V3(0.0f, 0.0f, 0.0f);
  }


}

void TMesh::Render(PPC *ppc, FrameBuffer *fb) {

  V3 *projVerts = new V3[vertsN];
  for (int vi = 0; vi < vertsN; vi++) {
    projVerts[vi] = V3(FLT_MAX, FLT_MAX, FLT_MAX);
    ppc->Project(verts[vi], projVerts[vi]);
  }

  for (int tri = 0; tri < trisN; tri++) {
    unsigned int vinds[3];
    vinds[0] = tris[3*tri+0];
    vinds[1] = tris[3*tri+1];
    vinds[2] = tris[3*tri+2];
    for (int ei = 0; ei < 3; ei++) {
      fb->DrawSegment(projVerts[vinds[ei]],
        projVerts[vinds[(ei+1)%3]],
        cols[vinds[ei]], cols[vinds[(ei+1)%3]]);
    }
  }

  delete []projVerts;

}

void TMesh::RotateAboutArbitraryAxis(V3 aO, 
  V3 aDir, float theta) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi].RotatePointAboutArbitraryAxis(aO, aDir, theta);
  }


}

V3 TMesh::GetCenter() {

  V3 ret(0.0f, 0.0f, 0.0f);
  for (int vi = 0; vi < vertsN; vi++) {
    ret = ret + verts[vi];
  }
  return ret/(float)vertsN;

}

void TMesh::Load(char *fname) {

  ifstream ifs(fname, ios::binary);
  if (ifs.fail()) {
    cerr << "INFO: cannot open file: " << fname << endl;
    return;
  }

  ifs.read((char*)&vertsN, sizeof(int));
  char yn;
  ifs.read(&yn, 1); // always xyz
  if (yn != 'y') {
    cerr << "INTERNAL ERROR: there should always be vertex xyz data" << endl;
    return;
  }
  if (verts)
    delete verts;
  verts = new V3[vertsN];

  ifs.read(&yn, 1); // cols 3 floats
  if (cols)
    delete cols;
  cols = 0;
  if (yn == 'y') {
    cols = new V3[vertsN];
  }

  ifs.read(&yn, 1); // normals 3 floats
  V3 *normals = 0; // don't have normals for now
  if (normals)
    delete normals;
  normals = 0;
  if (yn == 'y') {
    normals = new V3[vertsN];
  }

  ifs.read(&yn, 1); // texture coordinates 2 floats
  float *tcs = 0; // don't have texture coordinates for now
  if (tcs)
    delete tcs;
  tcs = 0;
  if (yn == 'y') {
    tcs = new float[vertsN*2];
  }

  ifs.read((char*)verts, vertsN*3*sizeof(float)); // load verts

  if (cols) {
    ifs.read((char*)cols, vertsN*3*sizeof(float)); // load cols
  }

  if (normals)
    ifs.read((char*)normals, vertsN*3*sizeof(float)); // load normals

  if (tcs)
    ifs.read((char*)tcs, vertsN*2*sizeof(float)); // load texture coordinates

  ifs.read((char*)&trisN, sizeof(int));
  if (tris)
    delete tris;
  tris = new unsigned int[trisN*3];
  ifs.read((char*)tris, trisN*3*sizeof(unsigned int)); // read tiangles

  ifs.close();

  cerr << "INFO: loaded " << vertsN << " verts, " << trisN << " tris from " << endl << "      " << fname << endl;
  cerr << "      xyz " << ((cols) ? "rgb " : "") << ((normals) ? "nxnynz " : "") << ((tcs) ? "tcstct " : "") << endl;

}

void TMesh::Translate(V3 transv) {

  for (int vi = 0; vi < vertsN; vi++) {
    verts[vi] = verts[vi] + transv;
  }

}
